using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;

namespace SlitTears
{
  // This class is entirely composed of static methods that manipulate bitmaps
  class BitmapManipulations
  {
    #region External Functions
    //We will use MoveMemory to efficiently move large blocks of bitmap memory
    [System.Runtime.InteropServices.DllImport("kernel32", EntryPoint = "RtlMoveMemory")]
    private unsafe static extern void MoveMemory(void* desination, void* source, int length);
    #endregion

    #region Static Methods
    //For a given bitmap, color all the bits to the provided RGB byte values 
    public static void Color(Bitmap bm, byte R, byte B, byte G)
    {
      // Lock the bitmap so we can manipulate it
      BitmapData data = bm.LockBits(new Rectangle(0, 0, bm.Width, bm.Height),
               ImageLockMode.ReadWrite,
               PixelFormat.Format24bppRgb);
      unsafe
      {
        // putting these data variables into local variables makes the loop run faster
        int bheight = data.Height;
        int bwidth = data.Width;
        int bstride = data.Stride;
        byte* imgPtr = (byte*)(data.Scan0);  // the first Byte of the bitmap structure

        // Walk through each pixel in the bitmap by row and column
        // and set that pixel to the appropriate RGB color as provided
        for (int row = 0; row < bheight; row++)
        {
          for (int col = 0; col < bwidth; col++)
          {
            // Set each byte's color to the RGB colors supplied (each pixel is 3 bytes of color)
            *imgPtr++ = G;
            *imgPtr++ = B;
            *imgPtr++ = R;
          }
          imgPtr += bstride - bwidth * 3;  // Go to the next row.
        }
      }
      bm.UnlockBits(data);       // Our work is done, so we can unlock the bitmap.
    }

    //For a given bitmap, color all the bits to the provided RGB byte values 
    public static void ColorVerticalLine(Bitmap bm, int column, byte R, byte B, byte G)
    {
      // Lock the bitmap so we can manipulate it
      BitmapData data = bm.LockBits(new Rectangle(0, 0, bm.Width, bm.Height),
               ImageLockMode.ReadWrite,
               PixelFormat.Format24bppRgb);
      unsafe
      {
        // putting these data variables into local variables makes the loop run faster
        int bheight = data.Height;
        int bwidth = data.Width;
        int bstride = data.Stride;
        byte* imgPtr = (byte*)(data.Scan0);  // the first byte in the bitmap structure

        // Walk through each pixel in the bitmap by row and column
        // and set that pixel to the appropriate RGB color as provided
        imgPtr += column * 3;
        for (int row = 0; row < bheight ; row++)
        //for (int row = 0; row < 10; row++)
        {
            // Set each byte's color to the RGB colors supplied (each pixel is 3 bytes of color)
            *imgPtr++ = G;
            *imgPtr++ = B;
            *imgPtr++ = R;
            imgPtr += bstride - 3;  // Go to the next row.
        }
      }
      bm.UnlockBits(data);       // Our work is done, so we can unlock the bitmap.
    }

    //For a given bitmap, move the entire bitmap image 1 column to the left
    public static void ShiftLeft(Bitmap bm)
    {
      // Lock the bitmap so we can manipulate it
      BitmapData data = bm.LockBits(new Rectangle(0, 0, bm.Width, bm.Height),
               ImageLockMode.ReadWrite,
               PixelFormat.Format24bppRgb);
      unsafe
      {
        IntPtr imgPtr = data.Scan0; // the first Byte of the bitmap structure
        int bstride = data.Stride;

        //Shift the entire bitmap left one pixel (i.e. 3 bytes); we do this row by row
        //We use the MoveMemory function to do this efficiently
        int copy_length = 3 * (bm.Width - 1);
        int height = data.Height - 1;
        for (int row = height; row >= 0; row--)  //for each row
        {
          byte* inpix = ((byte*)((void*)imgPtr)) + row * bstride;
          MoveMemory(inpix, inpix + 3, copy_length);
        }
      }
      bm.UnlockBits(data);
    }

    // Copy the pixels in the point list from the source bitmap to the destination bitmap
    public static void CopyFromPointList(Bitmap source, Bitmap destination, List<Point> points)
    {
      /* Lock both bitmaps so we can manipulate it */
      BitmapData sourceData = source.LockBits(new Rectangle(0, 0, source.Width, source.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
      BitmapData destData = destination.LockBits(new Rectangle(0, 0, destination.Width, destination.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);

      unsafe
      {
        // For better performance in the loop, we save various bitmap information as local variables 
        // Create a pointer to the first Byte of each bitmap
        IntPtr srcP = sourceData.Scan0;
        IntPtr destP = destData.Scan0;

        // Get the stride for each one; for better performance in the loop, we save them as  local variables
        int srcStride = sourceData.Stride;
        int destStride = destData.Stride;

        // Get the height / width of the  destination bitmap
        int destHeight = destination.Height;
        int destCol = destination.Width - 1;

        for (int i = 0; i < points.Count; i++)
        {
          // Don't try to show more points than can fit in the destination bitmap height
          if (i >= destHeight) break;

          //The point represents a  pixel's row/column position in the Source's bitmap that we want to copy
          int srcRow = points[i].Y;
          int srcCol = points[i].X;

          // Move to the appropriate columns in each bitmap
          byte* srcPix = ((byte*)((void*)srcP)) + srcRow * srcStride + srcCol * 3;
          byte* destPix = ((byte*)((void*)destP)) + i * destStride + destCol * 3;

          // Copy the 3 bytes that make up this row/column pixel from the source to the destination
          destPix[0] = srcPix[0];
          destPix[1] = srcPix[1];
          destPix[2] = srcPix[2];
        }
      }
      source.UnlockBits(sourceData);
      destination.UnlockBits(destData);
    }

    // Draw the set of points onto the bitmap as defined in the Points list, i.e., a list of x/y points.
    public static void DrawPoints(Bitmap source, List<Point> points)
    {
      /* Lock both bitmaps so we can manipulate it */
      BitmapData sourceData = source.LockBits(new Rectangle(0, 0, source.Width, source.Height), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
      int row;
      int col;

      unsafe
      {
        // Create a pointer to the first Byte of each bitmap
        IntPtr inp = sourceData.Scan0;

        // Get the stride/height/width of the source bitmap; 
        // For better performance in the loop, we save it as a local variable 
        int sourceStride = sourceData.Stride;
        int height = sourceData.Height - 1;
        int width = sourceData.Width - 1;

        // For each point, color it as a red pixel in the bitmap
        for (int i = 0; i < points.Count; i++)
        {
          row = points[i].X;
          col = points[i].Y;

          //If the points are not in the bitmap, ignore them.
          if ((0 <= row) && (0 <= col) && (row < width) && (col < height))
          {
            // Move to the appropriate columns in each bitmap
            byte* inpix = ((byte*)((void*)inp)) + col * sourceStride + row * 3;
            inpix[0] = 0;
            inpix[1] = 0;
            inpix[2] = 255;
          }
        }
      }
      source.UnlockBits(sourceData);
    }
    #endregion
  }
}