using System;
using System.Drawing;
using System.Windows.Forms;
using System.Drawing.Imaging; //Needed
using System.Collections.Generic;
using System.Collections;
using EasyImages;
using AviReader;
using AviWriter;

namespace SlitTears
{
  public partial class Form1 : Form
  {
    #region data structures
    // Delegates
    private delegate void SetPictureBoxImage(PictureBox pbox, Bitmap image);

    // Cameras and Bitmaps
    EasyImages.CameraClient camera;
    Bitmap bmSlitImage; //A bitmap that we will use to hold the slit-scanned image
    Bitmap bmPausedImage;

    //AVI files
    string aviFileName = ""; //Holds the name of the src AVI file to display
    AviReader.AviReader aviReader = new AviReader.AviReader();
    int aviFrameNumber = 0; //The frame we are currently looking at

    // Image history
    private ArrayList history = new ArrayList();
    int historyListSize = 0;

    // Flags
    bool scrubbing = false; //Whether we are scrubbing over the image
    bool pausing = false; // When we pause, we need to grab an image of the next frame, i.e., 
                          // we always pause on the frame after we click pause
    bool firstTime = true;
    // Data Structures 
    private List<Point> pointsConnected = new List<Point>();   //list of points making up lines
    private List<Point> pointsRubberBand = new List<Point>();   //list of points making up lines

    //enums
    enum lineTypes
    {
      Scribble,
      Line
    };
    lineTypes lineType = lineTypes.Line;

    # endregion

    #region Forms and Initialization
    public Form1()
    {
      InitializeComponent();
    }

    //As the GUI is loaded, create the camera, initialize the bitmaps and display things.
    private void Form1_Load(object sender, EventArgs e)
    {
      // the history list size and the frameslice size
      // will be the pixel width of the cubed pictureBox i.e., 1 frame / pixel column
      historyListSize = this.pbSlitImage.Width - 1;

      // Create and display blank black bitmap which will eventually hold the slit scanned image 
      bmSlitImage = new Bitmap(pbSlitImage.Width, pbSlitImage.Height);
      BitmapManipulations.Color((Bitmap)bmSlitImage, 0, 0, 0);
      DisplayImageInPictureBox(pbSlitImage, bmSlitImage);
      
      //Create a new camera object and set its initial properties
      camera = new EasyImages.CameraClient("DefaultCamera");
      camera.FramesPerSecond = 15;
      camera.ReceivedFrame += new EasyImages.CameraClient.ReceivedFrameEventHandler(camera_ReceivedFrame);
      camera.Start();
    }
    // When we are exiting, clenaup by stopping the camera and closing any opened files
    private void Form1_FormClosing(object sender, FormClosingEventArgs e)
    {
      camera.ReceivedFrame -= new EasyImages.CameraClient.ReceivedFrameEventHandler(camera_ReceivedFrame);
      camera.Stop();
      if (aviReader.Opened) aviReader.Close();
    }

    #endregion

    #region Camera callbacks
    // As we receive each frame...
    void camera_ReceivedFrame(object sender, EasyImages.CameraEventArgs e)
    {
      // A  hack. This autoscroll should be in the form load, but it doesn't work there.
      // Essentially, I need to scroll the panel to the right, 
      // but it only works after everything has appeared. Sigh.
      if (firstTime)
      {
        panel4.AutoScrollPosition = new Point(this.bmSlitImage.Width, 0);
        firstTime = false;
      }
      handleFrame(e.Bitmap);
    }
    #endregion

    #region Retrieving frames from files
    private void FrameFromFile()
    {
      if (rbPause.Checked && !pausing)
      {
        handleFrame(this.bmPausedImage); // the person has paused, so use the frozen image.
        return;
      }
      Bitmap bmap;
      if (this.aviFrameNumber < aviReader.Count)
      {
        this.Text = aviFrameNumber + "/" + aviReader.Count.ToString();
        //Get the current frame
        bmap = aviReader.ReadFrame(aviFrameNumber);
        handleFrame(bmap);
        aviFrameNumber += this.tbSkipFrames.Value; ;
      }
      else
      {
        aviFrameNumber = 0;
        FrameFromFile();
      }
    }
    #endregion

    #region bitmap manipulations and displays
    void handleFrame(Bitmap bmap)
    {
      if (this.rbPause.Checked && this.pausing)
      {
        // special case. When we click pause, we always want to store a clean copy of the next frame
        this.bmPausedImage = (Bitmap) bmap.Clone();
        this.pausing = false;
      }
      if (rbPlay.Checked)
      {
        if (this.pointsConnected.Count != 0)
        {
          // Shift the slit-scanned bitmap left by one column
          BitmapManipulations.ShiftLeft((Bitmap)this.bmSlitImage);

          // Copy the pixels at the given point locations in the source frame to the slit-scanned bitmap
          BitmapManipulations.CopyFromPointList(bmap, this.bmSlitImage, this.pointsConnected);

          // Draw the lines on the source bitmap
          BitmapManipulations.DrawPoints(bmap, this.pointsConnected);
        }
        if (0 != this.pointsRubberBand.Count)
        {
          BitmapManipulations.DrawPoints(bmap, this.pointsRubberBand);
        }

        // Display both bitmaps
        if (!this.scrubbing) DisplayImageInPictureBox(this.pbFrame, bmap);
        DisplayImageInPictureBox(this.pbSlitImage, this.bmSlitImage);
        // Add the frame to the history list. 
        historyAddBitmap(bmap);
      }
      else 
      {
        // We are paused; we still update the display so we can see line drawings,
        // but we use the last frame in the history list and we don't update the history
        // or the vertical slices. 
        Bitmap tmpBmap = (Bitmap)this.bmPausedImage.Clone();
        if (this.pointsConnected.Count != 0)
        {
          // Draw the lines on the history bitmap
          BitmapManipulations.DrawPoints(tmpBmap, this.pointsConnected);
        }
        if (0 != this.pointsRubberBand.Count)
        {
          BitmapManipulations.DrawPoints(tmpBmap, this.pointsRubberBand);
        }

        // Display both bitmaps
        if (!this.scrubbing) DisplayImageInPictureBox(this.pbFrame, tmpBmap);
      }
    }

    // Add the frame to the history list. 
    // If the list is larger than the useful size, remove the first frame.
    private void historyAddBitmap(Bitmap bmap)
    {
      this.history.Add(bmap);

      //keep enough of a history to fit the image
      if (this.history.Count > this.historyListSize)
      {
        this.history.RemoveAt(0);
      }
    }

    //Display the image in the picture box in the correct thread
    private void DisplayImageInPictureBox(PictureBox pbox, Image image)
    {
      if (pbox.InvokeRequired) // We are in the wrong thread. Call ourselves in the correct thread
      {
        SetPictureBoxImage theDelegate = new SetPictureBoxImage(DisplayImageInPictureBox);
        BeginInvoke(theDelegate, new object[] { pbox, image });
      }
      else // we are in the correct thread, so assign the image
      {
        pbox.Image = image;
        pbox.Refresh();
      }
    }

    #endregion

    #region pbFrame Mouse callbacks
    Point lastPoint = Point.Empty;
    private void pbFrame_MouseDown(object sender, MouseEventArgs e)
    {
      if (MouseButtons.Left != e.Button) return;

      // Never let the coords go outside the bitmap 
      lastPoint = InBoundsPoint(e.X, e.Y, pbFrame.Width, pbFrame.Height);
    }

    private void pbFrame_MouseMove(object sender, MouseEventArgs e)
    {
      if (MouseButtons.Left != e.Button) return;

      // Never let the coords go outside the bitmap 
      Point currentpoint = InBoundsPoint(e.X, e.Y, pbFrame.Width, pbFrame.Height);

      //Somehow we lost the mousedown event...
      if (Point.Empty == lastPoint)
      {
        lastPoint = currentpoint;
        return;
      }

      // unless the point is the same, draw the connecting line and reset the lastPoint
      if ((currentpoint.X != lastPoint.X) && (currentpoint.Y != (lastPoint.Y)))
      {
        drawLine(lastPoint, currentpoint, (lineType == lineTypes.Line));
        if (lineType == lineTypes.Scribble)
        {
          lastPoint = currentpoint;
        }
      }
    }

    private void pbFrame_MouseUp(object sender, MouseEventArgs e)
    {
      if (MouseButtons.Left != e.Button) return;

      // We want at least two different points before drawing a line or using it
      //if ((Point.Empty  == lastPoint) | (pointsConnected.Count == 0)) return;
      if (Point.Empty == lastPoint) return;

      // Never let the coords go outside the bitmap 
      Point currentpoint = InBoundsPoint(e.X, e.Y, pbFrame.Width, pbFrame.Height);

      if ((currentpoint.X != lastPoint.X) && (currentpoint.Y != lastPoint.Y))
      {
        drawLine(lastPoint, currentpoint, false);
      }
      lastPoint = Point.Empty;
    }

    private Point InBoundsPoint(int x, int y, int xmax, int ymax)
    {
      // Never let the coords go outside the bitmap 
      int newX = (x < 0) ? 0 : x;
      int newY = (y < 0) ? 0 : y;
      if (newX > xmax) newX = xmax;
      if (newY > ymax) newY = ymax;
      return (new Point(newX, newY));
    }

    #endregion

    #region pbSlitImage Callbacks
    // Set the flag for scrubbing mode
    private void pbSlitImage_MouseDown(object sender, MouseEventArgs e)
    {
      this.scrubbing = true;
    }

    private void pbSlitImage_MouseUp(object sender, MouseEventArgs e)
    {
      this.scrubbing = false;
    }

    private void pbSlitImage_MouseMove(object sender, MouseEventArgs e)
    {
      if (e.Button != MouseButtons.Left) return;
      int position = pbSlitImage.Width - e.X - 3; //Find the location of the slit in the bitmap
      position = this.history.Count - position;   //Find its corresponding location in the history list
      if (position < 0 || position >= history.Count) return; //Nothing at that spot; just exit

      //Draw a red line in the slit image... but to do this non-permanently, we have to clone it.
      //Its not the most elegant way to do it (as it can flicker) but it works.
      Bitmap tempBitmap = (Bitmap) this.bmSlitImage.Clone ();
      BitmapManipulations.ColorVerticalLine(tempBitmap, e.X - panel4.AutoScrollOffset.X, 255, 0, 0);
      DisplayImageInPictureBox(pbSlitImage, tempBitmap);

      DisplayImageInPictureBox(this.pbFrame, (Bitmap)this.history[position] );
    }
    #endregion

    #region Line drawing Algorithm

    // Bresenham's line drawing algorithm, adapted from http://www.cs.unc.edu/~mcmillan/comp136/Lecture6/Lines.html
    // Unless its a rubber band line, add the point coordinates determined by the algorithm to our array of points
    // comprising all line segments seen so far
    public void drawLine(Point p0, Point p1, bool rubberBandLine)
    {
      int x0 = p0.X;
      int y0 = p0.Y;
      int x1 = p1.X;
      int y1 = p1.Y;

      int dy = y1 - y0;
      int dx = x1 - x0;
      int stepx, stepy;

      pointsRubberBand.Clear();

      if (dy < 0) { dy = -dy; stepy = -1; } else { stepy = 1; }
      if (dx < 0) { dx = -dx; stepx = -1; } else { stepx = 1; }
      dy <<= 1;                                                  // dy is now 2*dy
      dx <<= 1;                                                  // dx is now 2*dx

      if (!rubberBandLine)
      {
        pointsConnected.Add(new Point(x0, y0));
      }
      else
      {
        pointsRubberBand.Add(new Point(x0, y0));
      }
      if (dx > dy)
      {
        int fraction = dy - (dx >> 1);                         // same as 2*dy - dx
        while (x0 != x1)
        {
          if (fraction >= 0)
          {
            y0 += stepy;
            fraction -= dx;                                // same as fraction -= 2*dx
          }
          x0 += stepx;
          fraction += dy;                                    // same as fraction -= 2*dy
          if (!rubberBandLine)
          {
            pointsConnected.Add(new Point(x0, y0));
          }
          else
          {
            pointsRubberBand.Add(new Point(x0, y0));
          }
        }
      }
      else
      {
        int fraction = dx - (dy >> 1);
        while (y0 != y1)
        {
          if (fraction >= 0)
          {
            x0 += stepx;
            fraction -= dy;
          }
          y0 += stepy;
          fraction += dx;
          if (!rubberBandLine)
          {
            pointsConnected.Add(new Point(x0, y0));
          }
          else
          {
            pointsRubberBand.Add(new Point(x0, y0));
          }
        }
      }
    }
    #endregion

    #region Interface Widget callbacks

    // Clear the pointlist, make the slitscan bitmap black, and display it
    private void btnClear_Click(object sender, EventArgs e)
    {
      pointsConnected.Clear();
      BitmapManipulations.Color((Bitmap)bmSlitImage, 0, 0, 0);
      DisplayImageInPictureBox(this.pbSlitImage, bmSlitImage);
      this.history.Clear();
      //if (this.rbPause.Checked && !pausing)
      //{
      //  handleFrame(this.bmPausedImage);
      //}
    }

    //Whenever the timer goes off, get the next frame from the file
    private void tmrFrameSpeed_Tick(object sender, EventArgs e)
    {
      FrameFromFile();
    }


    //Set the timer interval for reading files to the current value of this trackbar
    private void tbFrameSpeed_Scroll(object sender, EventArgs e)
    {
      tmrFrameSpeed.Interval = tbFrameSpeed.Value;
    }

    private void rbLine_CheckedChanged(object sender, EventArgs e)
    {
      lineType = lineTypes.Line;
    }

    private void rbScribble_CheckedChanged(object sender, EventArgs e)
    {
      lineType = lineTypes.Scribble;
    }

    //Toggle between generating frames from the camera vs. from a file
    private void cbUseFile_CheckedChanged(object sender, EventArgs e)
    {
      if (cbUseFile.Checked)
      {
        camera.Stop();
        if (openSrcAVIFileDialog.ShowDialog() == DialogResult.OK)
        {
          aviFileName = openSrcAVIFileDialog.FileName;
          aviReader.Open(aviFileName);
          tmrFrameSpeed.Interval = tbFrameSpeed.Value;
          tmrFrameSpeed.Enabled = true;
          pnlAVIFrameControl.Visible = true;
        }
      }
      else
      {
        aviReader.Close();
        tmrFrameSpeed.Enabled = false;
        pnlAVIFrameControl.Visible = false;
        camera.Start();
      }
    }

    private void rbPause_CheckedChanged(object sender, EventArgs e)
    {
      this.pausing = true;
    }

    // When we resize, auto scroll to the end of the bitmap so we can see the slices appear.
    private void Form1_ResizeEnd(object sender, EventArgs e)
    {
      panel4.AutoScrollPosition = new Point(this.bmSlitImage.Width, 0);
    }

    //Track bar to skip a # of frames while reading in a file (to make things faster)
    private void tbSkipFrames_Scroll(object sender, EventArgs e)
    {
      if (tbSkipFrames.Value == 1)
      {
        this.lblSkipFrames.Text = "Skip no frames";
      }
      else
      {
        this.lblSkipFrames.Text = "Skip every " + tbSkipFrames.Value.ToString() + " frames";
      }
    }
    #endregion
  }
}